const axios = require('axios');
const fs = require('fs');
const path = require('path');
const updaterProdutosModel = require('../model/updaterProdutosModel');
const updaterProdVersaoModel = require('../model/updaterProdVersaoModel');
const Filial = require('../model/Filial');
const UpdaterProdFili = require('../model/UpdaterProdFili');
const UpdaterProdFiliCx = require('../model/UpdaterProdFiliCx');
const SystemLogger = require('../logs/SystemLogger');
const AssistantControl = require('../service/AssistantControl');
const versionCheck = require('../functions/VersionCheck');
const Gzip = require('../functions/Gzip');
const Base64 = require('../functions/Base64');
const UpdaterProdConfig = require('../model/UpdaterProdConfig');

class ApiController {
  async insertUpdaterProdutos(product) {
    const params = {
      ID:          product.ID,
      NMPRODUCT:   product.NMPRODUTO,
      IDUTMULTVER: product.IDUTASSMULTVER,
    };
    return await updaterProdutosModel.insertUpdaterProdutos(params);
  }

  async updateUpdaterProdutos(id, product) {
    const params = {
      ID:          id,
      NMPRODUCT:   product.NMPRODUTO,
      IDUTMULTVER: product.IDUTASSMULTVER,
    };
    return await updaterProdutosModel.updateUpdaterProdutos(params);
  }

  async insertUpdaterProdVersao(versions, updaterProdId, actualVersionIds) {
    await Promise.all(versions.map(async (version) => {
      const fullVersion = `${version.VERSION}.${version.REVISION}.${version.BUILD}`;
      if(!actualVersionIds || !actualVersionIds.includes(fullVersion)) {
        const params = {
          UPDATERPRODID: updaterProdId,
          NRVERSAO:      fullVersion,
          DSURL:         version.URL,
          DSDIRETORIO:   version.PATH_TO_DEPLOY,
        };
        await updaterProdVersaoModel.insertUpdaterProdVersao(params);
      }
    }));
  }

  async insertUpdaterProdFili(units, updaterProdId, actualUnitIds) {
    await Promise.all(units.map(async (unit) => {
      if(!actualUnitIds || !actualUnitIds.includes(unit)) {
        const params = {
          UPDATERPRODID: updaterProdId,
          CDFILIAL:      unit
        };
        await UpdaterProdFili.insert(params);
      }
    }));
  }

  async downloadVersion(url, fullVersion, productPath, id) {
    const { PACKAGE_FOLDER } = process.env;
    const productVersionPath = `${PACKAGE_FOLDER}/${productPath}`;
    const versionPath = `${productVersionPath}/${fullVersion}.tar.gz`;
    let writer = null;

    try {
      if(!fs.existsSync(versionPath)) {
        if(!fs.existsSync(productVersionPath)) {
          fs.mkdirSync(productVersionPath, { recursive: true });
        }

        const writerPath = path.resolve(versionPath);
        writer = fs.createWriteStream(writerPath);

        const response = await axios({ url, method: 'GET', responseType: 'stream', });
        response.data.pipe(writer);

        writer.on('finish', async () => {
          await this.updateUpdaterProdVersao(id, 'S');
        });

        writer.on('error', async (error) => {
          await this.updateUpdaterProdVersao(id, 'E');
          SystemLogger.error(`ERRO CREATE VERSION FILE ${error.message}`);
        });
      }
    } catch (error) {
      if (writer) {
        writer.close();
      }

      await this.updateUpdaterProdVersao(id, 'E');
      SystemLogger.error(`ERROR DOWNLOAD VERSION ${error.message} URL: ${url}`);
      fs.unlink(versionPath, (err) => {
        if(err) {
          SystemLogger.error(`ERROR REMOVE DIRECTORY {${versionPath}} - ${err.message}`);
        }
      });
    }
  }

  async updateUpdaterProdVersao(id, status) {
    const params = {
      ID: id,
      IDDOWNLOADVER: status,
    };

    await updaterProdVersaoModel.updateUpdaterProdVersaoStatus(params);
  }

  getFilesFromPath(directoryPath) {
    if (!fs.existsSync(directoryPath)) {
      return [];
    }

    const files = fs.readdirSync(directoryPath);
    const filtredFiles = [];
    files.forEach((file) => {
      const filePath = path.join(directoryPath, file);
      const stat = fs.statSync(filePath);

      if (!stat.isDirectory() && file.endsWith('.tar.gz')) {
        // Recursively call listTarGzFiles for subdirectories
        filtredFiles.push(file);
      }
    });

    return filtredFiles;
  }

  loadLocalPackages(localPackages, DSDIRETORIO) {
    const { PACKAGE_FOLDER } = process.env;
    localPackages[DSDIRETORIO] = localPackages[DSDIRETORIO] || [];
    localPackages[DSDIRETORIO] = localPackages[DSDIRETORIO].concat(
      this.getFilesFromPath(path.join(PACKAGE_FOLDER, DSDIRETORIO))
    );
  }

  async handleProductConfig() {
    const productsConfig = await AssistantControl.getProductsConfig();

    if(!productsConfig) return null; // Error in integration with AssistantControl

    const mappedProductsConfig = productsConfig.reduce((acc, cv) => {
      acc[cv.devopsId] = cv.config;
      return acc;
    }, {});

    const productsIdWithConfig = await UpdaterProdConfig.getProductsWithConfig();
    productsIdWithConfig.rows.forEach(({ CDPRODUCTID }) => {
      delete mappedProductsConfig[CDPRODUCTID];
    });
    
    const productsWithoutConfig = Object.keys(mappedProductsConfig);
    await Promise.all(productsWithoutConfig.map(async (CDPRODUCTID) => {
      const decodedConfig = Gzip.decode(mappedProductsConfig[CDPRODUCTID]);
      return UpdaterProdConfig.deleteEmptyConfigByProduct({ CDPRODUCTID }).finally(() => {
        return UpdaterProdConfig.insert({
          CDPRODUCTID,
          NRVERSAO: '0.0.0',
          CONFIG: Base64.encode(decodedConfig),
        });
      });
    }));
  }

  async setLatestUpdaterApprovedVersion() {
    try {
      const controlUrl = AssistantControl.getControlUrl();
      const credentials = AssistantControl.getCredentials();

      const response = await axios.get(`${controlUrl}/updater-approved-version`, { auth: credentials }, { timeout: 15000 });

      process.env.UPDATE_AVAILABLE_VERSION = response.data.version || 'error';
    } catch (error) {
      SystemLogger.error(`ERRO BUSCAR VERSAO DISPONIVEL UPDATER: ${error.message}`);
      process.env.UPDATE_AVAILABLE_VERSION = 'error';
    }
  }

  async getLatestProductVersion(productId) {
    try {
      const versions = await updaterProdVersaoModel.getUpdaterProdVersaoByProduct(productId, true);

      const sortedVersions = versions.rows.sort((x, y) => {
        return versionCheck.sort(x.NRVERSAO, y.NRVERSAO);
      });

      const maxVersion = sortedVersions.shift();

      if (!('NRVERSAO' in maxVersion)) {
        throw 'Nenhuma versão do produto foi encontrada';
      }

      return maxVersion.NRVERSAO;
    } catch (error) {
      throw error;
    }
  }

  getLatestUpdaterApprovedVersion() {
    try {
      return process.env.UPDATE_AVAILABLE_VERSION;
    } catch (error) {
      return false;
    }
  }

  async checkRequisites(product, productInfo) {
    if (productInfo === null) {
      return {
        check: 'alert',
        message: 'Não foi possível estabelecer conexão com o Assistant Control'
      };
    }

    if (productInfo.length === 0) {
      return {
        check: 'valid',
      };
    }

    const { requiredProducts } = productInfo[0];

    if (requiredProducts.length === 0) {
      return {
        check: 'valid',
      };
    }

    const requiredProductsIds = requiredProducts.map(reqProd => {
      return reqProd.devopsId;
    });

    const params = {
      CDFILIAL: product.CDFILIAL,
      CDCAIXA: product.CDCAIXA,
      REQPRODS: requiredProductsIds.join(',')
    }

    const model = product.CDCAIXA ? UpdaterProdFiliCx : UpdaterProdFili;

    const missingReqProds = await model.getMissingRequiredProducts(params);
    
    if (missingReqProds.rows.length === 0) {
      return {
        check: 'valid',
      }
    }

    const result = {
      check: 'error',
      message: 'Produtos obrigatórios: '
    };

    const reqProdArr = [];

    for (const missingReq of missingReqProds.rows) {
      reqProdArr.push(missingReq.NMPRODUCT);
    }

    result.message += reqProdArr.join(' + ');

    return result;
  }

  async getOrgProducts() {
    const {
      NRORG,
      LICENSE_URL
    } = process.env;
    let result = { isError: false, response: null };

    try {
      const filter = {
        requestType : 'FilterData',
        filter : [ {
          'name' : 'NRORG', 'operator' : '=', 'value' : NRORG
        } ]
      };
      const axiosRep = await axios.post(`${LICENSE_URL}/getOrgProducts`, filter, { timeout: 60000 });

      if(axiosRep.data.dataset && axiosRep.data.dataset.products) {
        let licenseProductIds = [];
        const licenseProducts = axiosRep.data.dataset.products;
        let updaterProducts = await updaterProdutosModel.getUpdaterProdutos();
        updaterProducts = updaterProducts.rows;
        let { rows } = await Filial.getUnits();
        const units = rows.map((unit) => unit.CDFILIAL);

        //Insert/Update UPDATERPRODUTOS
        await Promise.all(licenseProducts.map(async (product) => {
          licenseProductIds.push(product.ID.toString());
          let auxProd = updaterProducts.find((item) => item.ID == product.ID);
          if(!auxProd) {
            await this.insertUpdaterProdutos(product);
            await this.insertUpdaterProdVersao(product.VERSIONS, product.ID);
            await this.insertUpdaterProdFili(units, product.ID);
          } else {
            await this.updateUpdaterProdutos(product.ID, product);

            const actualVersions = await updaterProdVersaoModel.getUpdaterProdVersao(product.ID);
            const actualVersionIds = actualVersions.rows.map((version) => { return version.NRVERSAO });
            await this.insertUpdaterProdVersao(product.VERSIONS, product.ID, actualVersionIds);

            const actualUnits = await UpdaterProdFili.getInstalledVersionsByProduct({ UPDATERPRODID: product.ID });
            const actualUnitIds = actualUnits.rows.map((unit) => { return unit.CDFILIAL });
            await this.insertUpdaterProdFili(units, product.ID, actualUnitIds);
          }
        }));

        //Cancel UPDATERPRODUTOS
        await Promise.all(updaterProducts.map(async (updaterProduct) => {
          if(!licenseProductIds.includes(updaterProduct.ID.toString())) {
            await updaterProdutosModel.inactiveUpdaterProdutos(updaterProduct.ID);
          }
        }));

        // Insert UPDATERPRODCONFIG for products without initial config
        await this.handleProductConfig().catch(error => SystemLogger.error(error));

        //Download all pending/error versions in background
        const localPackages = {};
        const versionToDownload = await updaterProdVersaoModel.getUpdaterProdVersaoToDownload();

        versionToDownload.rows.forEach(async (version) => {
          if (!localPackages[version.DSDIRETORIO]) {
            this.loadLocalPackages(localPackages, version.DSDIRETORIO);
          }

          let prodPackages = localPackages[version.DSDIRETORIO];

          if (!prodPackages || !prodPackages.includes(`${version.NRVERSAO}.tar.gz`)) {
            this.downloadVersion(version.DSURL, version.NRVERSAO, version.DSDIRETORIO, version.ID);
          } else if (version.IDDOWNLOADVER !== 'S') {
            await this.updateUpdaterProdVersao(version.ID, 'S');
          }
        });

        result.response = licenseProducts;
      } else {
        result.isError = true;
        if(axiosRep.data.error) {
          result.response = axiosRep.data.error;
        } else if(axiosRep.data.includes('Stack trace')) {
          result.response = axiosRep.data.split('Stack trace')[0];
        } else {
          result.response = axiosRep.data;
        }
      }
      return result;
    } catch (error) {
      result.isError = true;
      if(error.toJSON) {
        error = error.toJSON();
      }
      result.response = error.message;
      return result;
    }
  }
}

module.exports = new ApiController();
