const SystemLogger = require('../logs/SystemLogger');
const Telegram = require("../functions/Telegram");
const UpdaterService = require('../service/UpdaterService');
const UpdaterTelegramData = require('../service/UpdaterTelegramData');
const LogController = require('./LogController');
const { v4: uuidv4 } = require("uuid");
const UpdaterConfig = require('../model/UpdaterConfig');
const DB = require('../database/DB');
const Base64 = require('../functions/Base64');

class UpdaterControllerEx {

  async installVersion(req, res) {
    const data = req.body;
    data.ip = UpdaterService.getRequestIp(req);
    data.IDSTEVENTO = 'INSTALACAO_PENDENTE';

    const response = await UpdaterService.insertEventTable(data);
    UpdaterTelegramData.getData(data, 'single-install').then(telegramData => {
      Telegram.sendNotification('telegram-update-system', telegramData);
    });

    res.send(response);
  }

  async multiInstallVersion(req, res) {
    const connection = await DB.beginTransaction();
    let statusCode = 200;
    let result = {};
    try {
      const data = req.body;
      data.ip = UpdaterService.getRequestIp(req);
      data.IDSTEVENTO = 'INSTALACAO_PENDENTE';

      const { UNITS, ...info } = data;
      const insertEventsPromise = UNITS.map( unit => {
        return UpdaterService.insertEventTable({ ...info, ...unit }, connection);
      });

      UpdaterTelegramData.getData(data, 'multi-install').then(telegramData => {
        Telegram.sendNotification('telegram-update-system', telegramData);
      });
      const response = await Promise.all(insertEventsPromise);
      connection.commit();
      result = { INSTALLID: uuidv4(), result: response };
    } catch(err) {
      statusCode = 500;
      result = {INSTALLID: null, result: [], error: true };
      connection.rollback();
    } finally {
      connection.close();
    }
    res.status(statusCode).send(result);
  }

  async removeVersion(req, res) {
    const data = req.body;
    data.ip = UpdaterService.getRequestIp(req);
    data.IDSTEVENTO = 'REMOCAO_PENDENTE';

    const response = await UpdaterService.insertEventTable(data);

    UpdaterTelegramData.getData(data, 'single-install').then(telegramData => {
      Telegram.sendNotification('telegram-update-system', telegramData);
    });

    res.send(response);
  }

  async installVersionInProgress(req, res) {
    const data = req.body;

    const response = await UpdaterService.updateEventTable(data);

    res.send(response);
  }

  async removeVersionInProgress(req, res) {
    const data = req.body;

    const response = await UpdaterService.updateEventTable(data);

    res.send(response);
  }

  async finishUpdater(req, res) {
    const data = req.body;

    const response = await UpdaterService.finishEventTable(data);
    res.send(response);
  }

  async cancelUpdater(req, res) {
    const data = req.body;

    const response = await UpdaterService.cancelEventTable(data);

    res.send(response);
  }

  async cancelUnitEvents(req, res) {
    const data = req.body;

    await UpdaterService.cancelUnitEvents(data)

    const response = {
      error: false,
      message: `Os eventos da filial ${data.CDFILIAL} foram cancelados.`
    }

    res.send(response);
  }

  async getPendingEventsByUnit(req, res) {
    const data = req.body;

    const response = await UpdaterService.getPendingEventsByUnit(data.CDFILIAL, data.CDCAIXA);

    res.send(response);
  }

  async getAvailablePackages(req, res) {
    const data = req.body;
    const extension = 'tar.gz';
    const { PACKAGE_SERVER } = process.env;
    const packageURL = `${PACKAGE_SERVER}/${data.DIRECTORY}/${data.INSTALLVERSION}.${extension}`

    res.send({url: packageURL});
  }

  async checkInstalledVersion(req, res){
    const params = req.body;
    const checkVersionInstall = await UpdaterService.checkVersion(params);
    res.send(checkVersionInstall);
  }

  async saveConfigDb(req, res){
    const params = req.body;

    const response = await UpdaterService.saveConfig(params);

    res.send(response);
  }

  async getConfigIniByUnit(req, res) {
    const { CDFILIAL: P_CDFILIAL } = req.body;

    const response = await UpdaterConfig.getConfigIni({ P_CDFILIAL });
    if(!response.rows.length) {
      res.status(500).send(`Não existe um arquivo de configuração salvo para a filial ${P_CDFILIAL}`);
      return;
    }
    const [ firstRow ] = response.rows;
    const data = Base64.decode(firstRow.CONFIG);
    res.send({ data });
  }

  async saveConfigIniByUnit(req, res) {
    const {
      CDFILIAL,
      CDOPERADOR,
      CONFIG
    } = req.body;

    const response = await UpdaterService.saveConfigIniByUnit(CDFILIAL, CDOPERADOR, CONFIG);

    res.send(response);
  }

  async getConfigHistoryByUnit(req, res) {
    const params = req.body;

    const response = await UpdaterService.getConfigHistoryByUnit(params.CDFILIAL);

    res.send(response);
  }

  async getProdConfig(req, res){
    const params = req.body;

    const response = await UpdaterService.getProdConfig(params);

    res.send(response);
  }

  async saveProdConfig(req, res){
    const params = req.body;

    const response = await UpdaterService.saveProdConfig(params);

    res.send(response);
  }

  async installNewVersion(req, res) {
    const params = req.body;

    const response = await UpdaterService.installNewVersion(params);

    res.send(response)
}

  async removeEventsByUnitAndProduct(req, res) {
    const params = req.body;

    const response = await UpdaterService.removeEventsByUnitAndProduct(params);

    res.send(response);
  }

  async updatePackageURL(req, res) {
    const params = req.body;

    const response = await UpdaterService.updatePackageURL(params);

    res.send(response);
  }

  async uploadLog(req, res) {
    try {
      const { file, type, param } = req.body;
      if(type === 'zip') {
        await LogController.setLogZipFile(file, param);
      } else if(type === 'txt') {
        await LogController.setUpdaterLog(file, param);
      }
      res.send({ error: false, message: 'SUCCESS_LOADED_LOG' });
    } catch (error) {
      SystemLogger.error(error);
      res.send({
        error: true,
        message: error.message
      });
    }
  }

  async getLogFile(req, res){
    const logFile = await LogController.getUpdaterLog();
    res.on('finish', () => {
      LogController.clearLogFile();
    });
    res.send(logFile);
  }

  async verifyInstalledProducts(req, res) {
    const { CDFILIAL, CDOPERADOR } = req.body;
    const data = {
      CDFILIAL,
      CDOPERADOR,
      IDSTEVENTO: 'VERIFICACAO_PENDENTE',
    };
    const response = await UpdaterService.insertEventTable(data);

    res.send(response);
  }

  async verifyIntegrity(req, res) {
    try {
      const { CDFILIAL, CDOPERADOR, PRODVERSIONS } = req.body;
      if(Array.isArray(PRODVERSIONS)) {
        const resultPromiseArr = PRODVERSIONS.map(({ PRODFILIID }) => {
          return UpdaterService.updateAnalyzeStatus(PRODFILIID, 'PENDENTE');
        });

        await Promise.all(resultPromiseArr);

        const data = {
          CDFILIAL,
          CDOPERADOR,
          IDSTEVENTO: 'ANALISE_INTEGRIDADE_PENDENTE',
        };

        const response = await UpdaterService.insertEventTable(data);
        return res.send(response);
      }
      throw new Error('Invalid product versions');
    } catch (err) {
      res.send({
        error: true,
        message: err.message,
      });
    }
  }

}

module.exports = new UpdaterControllerEx()