const ms = require('ms');
const monitoringController = require("../controller/MonitoringController");
const AssistantConfig = require("../model/AssistantConfig");
const UpdaterServicoFili = require("../model/UpdaterServicoFili");
const SocketRooms = require("./SocketRooms");
const ServicesMonitoring = require("../helpers/ServicesMonitoring");
const Gzip = require('../functions/Gzip');
const SystemLogger = require("../logs/SystemLogger");
const Filial = require("../model/Filial");

class MonitoringSocket {
    tenMinutes = ms('10m');
    monitoringParams = {};
    monitoringIntervals = {};
    monitoringResult = {};
    unitsWithSat = new Set();
    intervalLoadUnitsWithSat = null;
    timeLastConnectedUser = 0;
    intervalControlMonitoring = setInterval(() => {
        const hasActiveMonitoring = !!Object.keys(this.monitoringIntervals).length;
        const hasActiveUser = this.timeLastConnectedUser === null;
        const timeLapseLastUser = Date.now() - (this.timeLastConnectedUser || Date.now());
        if(hasActiveMonitoring && !hasActiveUser && timeLapseLastUser >= this.tenMinutes) {
            this.clearAllIntervals();
            this.monitoringIntervals = {};
        }
    }, this.tenMinutes);
    joinDefaultRoom( socket ) {
        socket.join(SocketRooms.client.default);
    }
    joinRooms( socket ) {
        socket.join(SocketRooms.master.default);
        socket.join(SocketRooms.master.monitoringPage);
    }
    async hasConsumersForMonitoring(){
        const consumers = await this.getRoom(SocketRooms.master.monitoringPage).fetchSockets();
        return consumers.length > 0;
    }
    async loadInitialParams(){
        return Promise.all([
            this.loadUnitsWithSat(),
            this.loadMonitoringParams(),
        ])
    }
    async connectedClient(UNITS = null, STATUS = 'S') {
        UNITS = Array.isArray(UNITS) ? UNITS : await this.getClients().then( sockets => {
            return sockets.map( socket => socket.CDFILIAL);
        });
        this.getRoom(SocketRooms.master.monitoringPage).emit("login", {STATUS, UNITS});
    }
    connectedUser(){
        this.timeLastConnectedUser = null;
    }
    async disconnectedUser() {
        const usersWithMonitoringActive = await this.getRoom(SocketRooms.master.monitoringPage).fetchSockets();
        if(usersWithMonitoringActive.length === 0) {
            this.timeLastConnectedUser = Date.now();
        }
    }
    clearAllIntervals(){
        this.getAllIntervals().forEach(clearInterval);
    }
    getAllIntervals(){
        let intervals = [];
        Object.keys(this.monitoringIntervals).forEach( CDFILIAL => {
            intervals = intervals.concat(Object.values(this.monitoringIntervals[CDFILIAL]));
        });
        return intervals;
    }
    getConfigById = (ID, paramsToInclude = {}) => {
        return ({...this.monitoringParams[ID], ...paramsToInclude});
    }
    getNewUnitsWithSat(updatedUnits){
        return [...updatedUnits].filter(unit => !this.unitsWithSat.has(unit));
    }
    async setUnitsWithSat( unitsWithSat, updateMonitoring = false ) {
        if(updateMonitoring) {
            const newUnitsWithSat = this.getNewUnitsWithSat(unitsWithSat);
            if(newUnitsWithSat.length) {
                const clientSockets = await this.getSocketsClientByUnits(newUnitsWithSat);
                const satMonitoring = this.getMonitorsParams().find(config => config.COMANDO === 'INTERNO' && config.NOME === 'SAT');
                this.unitsWithSat = unitsWithSat;
                this.startMonitoring(clientSockets, satMonitoring.ID);
            }
        } else {
            this.unitsWithSat = unitsWithSat;
        }

    }
    async fetchUnitsWithSat(){
        const queryResultAll = await Filial.getUnits();
        const { rows } = queryResultAll;
        const unitsWithSat = new Set();
        rows.forEach(({SAT, CDFILIAL}) => {
            // Tratamento especial para o Sat: S => Sucesso | E => Error | N => Filial não utiliza SAT
            if( SAT !== 0 ) unitsWithSat.add(CDFILIAL);
        });
        return unitsWithSat;
    }
    async loadUnitsWithSat() {
        this.setUnitsWithSat(await this.fetchUnitsWithSat());
        this.intervalLoadUnitsWithSat = setInterval(() => {
            this.fetchUnitsWithSat().then( units => this.setUnitsWithSat(units, true) );
        }, ms('1d'));
    }
    isUnitWithSat( CDFILIAL ) {
        return this.unitsWithSat.has(CDFILIAL);
    }
    stopMonitoring( ID ) {
        Object.values(this.monitoringIntervals).forEach( unit => {
            if(unit[ID]) clearInterval(unit[ID]);
        })
    }

    restartMonitoring( ID ) {
        this.startMonitoring(null, ID);
    }

    getMonitorsParams(ID = null) {
        if( !ID ) return Object.values(this.monitoringParams);
        if( this.monitoringParams[ID] ) return [this.monitoringParams[ID]];
        return [];
    }

    getUnitBySocketClient( socket ) {
        const { headers = {} } = socket.handshake;
        const { cdfilial = 'Unidade não informada no socket-client' } = headers;
        return cdfilial;
    }

    isMonitoringServiceInUnit( ID = null ) {
        const {
            TIPO,
            NOME
        } = this.monitoringParams[ID] || {};
        return (ID && TIPO === 'I' && NOME === 'Serviços Updater');
    }

    hasActiveMonitoring() {
        const intervalsUnits = Object.values(this.monitoringIntervals);
        const activeMonitoring = intervalsUnits.find(intervalsUnit => intervalsUnit && Object.keys(intervalsUnit).length);
        return !!activeMonitoring;
    }

    async startMonitoring( clients = null, ID = null ){
        const hasActiveMonitoring = this.hasActiveMonitoring();
        const isNewClientConnection = !!clients;
        const isUpdateMonitoringParams = !!ID;
        const needCreateOrUpdateMonitoring = isNewClientConnection
                                          || isUpdateMonitoringParams
                                          || !hasActiveMonitoring;
        if(!needCreateOrUpdateMonitoring) {
            const allClients = await this.getClients();
            const clientsWithoutMonitoring = allClients.filter(socket => {
                return !(socket.CDFILIAL in this.monitoringIntervals);
            });
            if(!clientsWithoutMonitoring.length) {
                return;
            }
            clients = clientsWithoutMonitoring;
        }
        clients = Array.isArray(clients) ? clients : await this.getClients();
        clients.forEach((socketClient) => {
            let monitorsToExecute = this.getMonitorsParams(ID);
            monitorsToExecute.forEach(config => {
                const ID = config.ID;
                const CDFILIAL = this.getUnitBySocketClient(socketClient);
                //? A => Execução na base do Assistant
                //? L => Execução na base Lojas, envia a requisição e aguarda o retorno para emitir para o Assistant
                //? I => Execução de rotinas criadas internamente que não podem ser alteradas via sistema (SAT, por exemplo)
                //? E => Execução de requisição HTTP nas Lojas, envia a requisição e aguarda o retorno para emitir para o Assistant
                //? S => Verificação se Serviços do Windows na Loja estão em execução
                const functionPerType = {
                    A: this.monitoringDataBaseCloud,
                    L: this.monitoringDataBaseLoja,
                    I: this.monitoringAssistantInterno,
                    E: this.monitoringApacheLoja,
                    S: this.monitoringServicesLoja,
                };
                const functionToCall = functionPerType[config.TIPO] || (() => {});
                let conf = this.getConfigById(ID, {CDFILIAL});
                functionToCall.apply(this, [conf, socketClient]);
                this.monitoringIntervals[CDFILIAL] = this.monitoringIntervals[CDFILIAL] || {};
                let unitIntervals = this.monitoringIntervals[CDFILIAL];
                if(unitIntervals[config.ID]) { // Parametros do monitoramento alterados interval deve ser recriado
                    clearInterval(unitIntervals[config.ID]);
                    unitIntervals[config.ID] = null;
                }
                unitIntervals[config.ID] = unitIntervals[config.ID] || setInterval(() => { // Não recria interval caso seja apenas uma nova conexão
                    let conf = this.getConfigById(ID, {CDFILIAL});
                    functionToCall.apply(this, [conf, socketClient]);
                }, config.INTERVALO * 1000);
            })
        });
    }
    hasResult( dataset ) {
        return dataset && dataset.rows && dataset.rows.length;
    }
    async loadMonitoringParamsServicesInUnit( ID ){
        const updaterServicoFili = await UpdaterServicoFili.getServiceMonitoring();
        if(ID && this.hasResult(updaterServicoFili)) {
            const servicesInUnits = updaterServicoFili.rows;
            this.monitoringParams[ID].UNITS = servicesInUnits.reduce((acc, cv) => {
                acc[cv.CDFILIAL] = {COMANDO: cv.MONITORAMENTO};
                return acc;
            }, {});
        }
    }
    async loadMonitoringParams(){
        try {
            const monitoringParams = await AssistantConfig.getMonitoringParams();

            let idServicoFiliMonitor = null;
            if(this.hasResult(monitoringParams)) {
                this.monitoringParams = monitoringParams.rows.reduce((acc, cv) => {
                    acc[cv.ID] = {...cv};
                    if(cv.TIPO === 'I' && cv.NOME === 'Serviços Updater') idServicoFiliMonitor = cv.ID;
                    if(cv.TIPO !== 'I' && cv.COMANDO) {
                        acc[cv.ID].COMANDO = Gzip.decode(cv.COMANDO);
                    }
                    return acc;
                }, {});
            }
            await this.loadMonitoringParamsServicesInUnit(idServicoFiliMonitor);
            this.startMonitoring();
        } catch (error) {
            return Promise.reject(error);
        }
    }

    async getSocketsClient(){
        return await this.getSocketIo().in("ASSISTANT_CLIENT_ROOM").fetchSockets();
    }

    async getSocketClientByUnit(CDFILIAL) {
        return this.getSocketsClient()
            .then((allSockets) => {
                const clientUnitSocket = allSockets.find((socket) => {
                    return socket.CDFILIAL == CDFILIAL;
                });
                return clientUnitSocket;
            });
    }
    async getSocketsClientByUnits( UNITS ) {
        const sockets = await this.getSocketsClient();
        return sockets.filter( socket => UNITS.includes(socket.CDFILIAL));
    }
    setSocketIo(io) {
        this.io = io;
    }

    getSocketIo() {
        return this.io;
    }

    getRoom(roomName) {
        return this.getSocketIo().to(roomName);
    }

    async getClients() {
        return await this.getRoom(SocketRooms.client.default).fetchSockets();
    }
    updateMonitoringParams(config) {
        const {ATIVO, ...monitorParams} = config;
        if(ATIVO == 'N') {
            return this.stopMonitoring(config.ID);
        }
        if(config.TIPO === 'S') {
            monitorParams.COMANDO = ServicesMonitoring.removeMonitoringParamInconsistencies(monitorParams.COMANDO);
        }
        this.monitoringParams[config.ID] = this.monitoringParams[config.ID] || {};
        Object.assign(this.monitoringParams[config.ID], monitorParams);
        this.hasConsumersForMonitoring().then(hasConsummers => {
            if(hasConsummers || this.hasActiveMonitoring()) {
                this.startMonitoring(null, config.ID);
            }
        });
    }
    registerMasterEvents(socket) {
        socket.on("getLatestMonitoringResult", () => {
            const result = Object.keys(this.monitoringResult).reduce((response, CDFILIAL) => {
                response[CDFILIAL] = {};
                Object.keys(this.monitoringResult[CDFILIAL]).forEach(MONITOR_ID => {
                    const { STATUS, QTDERROR } = this.monitoringResult[CDFILIAL][MONITOR_ID];
                    response[CDFILIAL][MONITOR_ID] = STATUS;
                    response[CDFILIAL].QTDERROR = response[CDFILIAL].QTDERROR || {};
                    response[CDFILIAL].QTDERROR[MONITOR_ID] = QTDERROR.toString();
                });
                return response;
            }, {});
            socket.emit('monitoringResultSummary', result);
        });

        socket.on("detailMonitoring", (config) => {
            const detail = this.monitoringResult[config.CDFILIAL]
                ? this.monitoringResult[config.CDFILIAL][config.ID]
                : [];
            socket.emit("detailMonitoringResult", detail);
        });
    }

    registerClientEvents( socket ) {
        socket.on(`MONITORING_RESULT_APACHE`, (result) => {
            try {
                const { ID, NOME, CDFILIAL } = result.data;
                const config = { ID, NOME, CDFILIAL };
                if (!result.isError) {
                    const isError = result.response.rows.some((element) => {
                        return element.STATUS == "N";
                    });
                    if (!isError) {
                        result.response.rows = [];
                    }
                    // console.log(
                    //     `Filial: ${CDFILIAL} | HTTP Loja SUC: ${ID} - ${NOME}`
                    // );
                    this.setMonitoringResult(config, result.response);
                } else {
                    const response = {
                        metaData: [{ name: "ERRO" }],
                        rows: [[result.response]],
                    };
                    SystemLogger.error(`Filial:', ${CDFILIAL} | HTTP Loja ERR: ${ID} - ${NOME} | Message: ${result.response}`);
                    this.setMonitoringResult(config, response);
                }
            } catch( error ) {
                SystemLogger.error(`OCCURRED_ERROR_APACHE_RESULT: ${error.message}`);
            }
        });

        socket.on(`MONITORING_RESULT_LOCAL_DB`, (result) => {
            try {
                const { ID, NOME = '', CDFILIAL, TIPO, COMPLEMENTOTIPO } = result.data;
                const config = { ID, NOME, CDFILIAL, TIPO, COMPLEMENTOTIPO };
                if (!result.isError) {
                    // console.log(`Filial: ${CDFILIAL} | SQL Loja SUC: ${ID} - ${NOME} | Quantidade: ${result.response.rows.length}`);
                    this.setMonitoringResult(config, result.response);
                } else {
                    const response = {
                        metaData: [{ name: "ERRO" }],
                        rows: [[result.response]],
                    };
                    SystemLogger.error(`Filial: ${CDFILIAL} | SQL Loja ERR: ${ID} - ${NOME} | Messsage: ${result.response}`);
                    this.setMonitoringResult(config, response);
                }
            } catch( error ) {
                SystemLogger.error(`OCCURRED_ERROR_LOCAL_DB_RESULT: ${error.message}`);
            }
        });

        socket.on(`MONITORING_RESULT_LOCAL_SERVICES`, (result) => {
            try {
                const { ID, NOME, CDFILIAL } = result.data;
                const config = { ID, NOME, CDFILIAL };
                if (!result.isError) {
                    const isError = result.response.rows.some((element) => {
                        return element.STATUS == "N";
                    });
                    if (!isError) {
                        result.response.rows = [];
                    }
                    // console.log(
                    //     "Filial:",
                    //     CDFILIAL,
                    //     "| Services Loja SUC:",
                    //     ID + " - " + NOME
                    // );
                    this.setMonitoringResult(config, result.response);
                } else {
                    const response = {
                        metaData: [{ name: "ERRO" }],
                        rows: [[result.response]],
                    };
                    SystemLogger.error(`Filial: ${CDFILIAL} | Servicos Loja ERR ${ID} - ${NOME} | Message: ${result.response}`);
                    this.setMonitoringResult(config, response);
                }
            } catch( error ) {
                SystemLogger.error(`OCCURRED_ERROR_SERVICES_RESULT: ${error.message}`);
            }
        });

        socket.on("MONITORING_RESULT_SAT", (result) => {
            try {
                const { ID, NOME, CDFILIAL } = result.data;
                const config = { ID, NOME, CDFILIAL };
                if (!result.isError) {
                    const isSatError = result.response.rows.some((element) => {
                        return element.sat === 'E';
                    });
                    if (!isSatError) {
                        result.response.rows = [];
                    }
                    // console.log(
                    //     `Filial: ${CDFILIAL} | Exec SAT SUC: ${ID} - ${NOME} | Error: ${isSatError}`
                    // );
                    this.setMonitoringResult(config, result.response);
                } else {
                    const response = {
                        metaData: [{ name: "ERRO" }],
                        rows: [[result.response]],
                    };
                    SystemLogger.error(`Filial: ${CDFILIAL} | SAT ERR: ${ID} - ${NOME} | Message: ${result.repsonse}`);
                    this.setMonitoringResult(config, response);
                }
            } catch( error ) {
                SystemLogger.error(`OCCURRED_ERROR_SAT_RESULT: ${error.message}`);
            }
        });
    }
    registerEvents(socket, agent) {
        if(agent === 'monitor') {
            this.registerClientEvents(socket);
        } else if(agent !== 'updater') {
            this.registerMasterEvents(socket);
        }
    }
    monitoringDataBaseCloud(config) {
        const { ID, NOME, CDFILIAL } = config;
        monitoringController
            .execCommand(config.COMANDO, { CDFILIAL })
            .then((response) => {
                // console.log(
                //     `Filial: ${CDFILIAL} | DB Cloud SUC: ${ID} - ${NOME} | Quantidade: ${response.rows.length}`
                // );
                this.setMonitoringResult(config, response);
            })
            .catch((error) => {
                const response = {
                    metaData: [{ name: "ERRO" }],
                    rows: [[error.message]],
                };
                SystemLogger.error(`Filial: ${CDFILIAL} | Cloud DB ERR: ${ID} - ${NOME} | Message: ${error.message}`);
                this.setMonitoringResult(config, response);
            });
    }
    setMonitoringResult(config, response) {
        //Atualiza a lista a ser usada no grid de monitoramento.
        const { CDFILIAL, ID, TIPO, COMPLEMENTOTIPO } = config;
        //Atualiza os dados do modal de detalhes de cada monitoramento.
        this.monitoringResult[CDFILIAL] = this.monitoringResult[CDFILIAL] || {};
        this.monitoringResult[CDFILIAL][ID] = response;
        const active = response.rows.every( item => item.STATUS === 'S');
        let QTDERROR = 0;
        let STATUS = '';

        for(const item of response.rows) {
            if(!item.STATUS) {
                QTDERROR = response.rows.length;
                break;
            }
            if(item.STATUS === 'N') QTDERROR++;
        };

        switch (TIPO) {
            case 'A':
            case 'L':
                if (COMPLEMENTOTIPO === 'S') {
                    STATUS = response.rows.length > 0 ? 'S' : 'N';
                } else {
                    STATUS = response.rows.length > 0 ? 'N' : 'S';
                }
                break;
            default:
                STATUS = active ? 'S' : 'N';
        }
        if(STATUS === 'S') QTDERROR = 0;
        this.monitoringResult[CDFILIAL][ID].STATUS = STATUS;
        this.monitoringResult[CDFILIAL][ID].QTDERROR = QTDERROR;
        //Envia informações atualizadas para a tela de monitoramento.
        this.getRoom("ASSISTANT_MASTER_MONITORING_ROOM").emit(ID, {ID, CDFILIAL, STATUS, QTDERROR});
    }

    monitoringDataBaseLoja(config, socketClient) {
        //Envia para a Loja
        socketClient.emit("execMonitoringSQL", config);

        //Recebe a resposta da Loja, sempre atualizando o listener para não duplicar
    }

    monitoringApacheLoja(config, socketClient) {
        //Envia para a Loja
        socketClient.emit("execMonitoringHttp", config);
        //Recebe a resposta da Loja, sempre atualizando o listener para não duplicar
    }

    monitoringServicesLoja(config, socketClient) {
        //Envia para a Loja
        socketClient.emit("execMonitoringServices", config);

        //Recebe a resposta da Loja, sempre atualizando o listener para não duplicar
    }
    monitoringSat(config, socketClient) {
        const { CDFILIAL } = config;
        const unitWithSat = this.isUnitWithSat(CDFILIAL);
        const monitorsByUnit = this.monitoringIntervals[CDFILIAL];
        const satIntervalInUnit = (monitorsByUnit || {})[config.ID];
        if(unitWithSat)            socketClient.emit("satMonitoring", config);
        else if(satIntervalInUnit) clearInterval(satIntervalInUnit);
    }
    monitoringServicesByUnit(config, socketClient) {
        const { ID, CDFILIAL, UNITS = {}, ...configMonitor } = config;
        const { COMANDO } = (UNITS[CDFILIAL] || {});
        if(ID && COMANDO) {
            const newConfig = {...configMonitor, COMANDO, CDFILIAL, ID};
            this.monitoringServicesLoja(newConfig, socketClient);
        }
    }
    monitoringAssistantInterno(config, socketClient) {
        switch( config.NOME ) {
            case 'SAT':
                this.monitoringSat(config, socketClient);
                break;
            case 'Serviços Updater':
                this.monitoringServicesByUnit(config, socketClient);
                break;
            default:
                SystemLogger.error(`INVALID_INTERNAL_MONITORING: ${config.NOME}`);
        }
    }
    getServicesUpdaterMonitor( CDFILIAL ){
        let { ID, UNITS } = Object.values(this.monitoringParams).find( monitor => {
            return monitor.TIPO === 'I' && monitor.NOME === 'Serviços Updater';
        }) || {ID: null, UNITS: {}};
        if(ID && !UNITS) {
            this.monitoringParams[ID].UNITS = {};
            UNITS = this.monitoringParams[ID].UNITS;
        }
        UNITS[CDFILIAL] = UNITS[CDFILIAL] || {COMANDO: ''};
        const { COMANDO } = UNITS[CDFILIAL];
        return {
            ID,
            COMANDO,
            PARAMS: UNITS[CDFILIAL]
        };
    }
    updateServiceToMonitoring(CDFILIAL, services, operation) {
        const { ID, COMANDO, PARAMS } = this.getServicesUpdaterMonitor( CDFILIAL );
        if( !ID ) return;
        let actualServices = COMANDO.length ? new Set(COMANDO.split(';')) : new Set();
        const addService = service => actualServices.add(service);
        const deleteService = service => actualServices.delete(service);
        services.split(';').forEach(operation === 'add' ? addService : deleteService);
        PARAMS.COMANDO = [...actualServices].join(';');
        return { ID, COMANDO: PARAMS.COMANDO, UPDATED: COMANDO !== PARAMS.COMANDO };
    }
    async addServiceToMonitoring(CDFILIAL, services) {
        try {
            const { ID, UPDATED } = this.updateServiceToMonitoring(CDFILIAL, services, 'add') || {};
            if(!ID) return;
            const monitorsByUnit = this.monitoringIntervals[CDFILIAL] || {};
            if( UPDATED && !monitorsByUnit[ID] ) {
                const socketClient = await this.getSocketClientByUnit(CDFILIAL);
                if(socketClient) {
                    this.startMonitoring([socketClient], ID);
                }
            }
        } catch( error ) {
            SystemLogger.error(`ERR_ADD_MONITORING: ${error.message}`);
        }
    }
    removeServiceToMonitoring(CDFILIAL, services) {
        try {
            const { ID, COMANDO, UPDATED } = this.updateServiceToMonitoring(CDFILIAL, services, 'delete') || {};
            if(!ID) return;
            const monitorsByUnit = this.monitoringIntervals[CDFILIAL] || {};
            if(UPDATED && !COMANDO && monitorsByUnit && monitorsByUnit[ID]) {
                clearInterval(monitorsByUnit[ID]);
                delete monitorsByUnit[ID];
                let monitoringResultByUnit = (this.monitoringResult[CDFILIAL] || {});
                if(monitoringResultByUnit[ID]) {
                    monitoringResultByUnit[ID] = false;
                }
                this.getRoom(SocketRooms.master.monitoringPage).emit(ID, {
                    ID,
                    CDFILIAL,
                    STATUS: false
                })
            }
        } catch( error ) {
            SystemLogger.error(`ERR_REMOVE_MONITORING: ${error.message}`);
        }
    }
}

module.exports = new MonitoringSocket();